###############################################################################
#                                                                             #
# SalesforceConnector                                                         #
# Copyright(C), 2006, Plone Foundation - ALL RIGHTS RESERVED                  #
#                                                                             #
# This software is licensed under the Terms and Conditions contained within   #
# the License.txt file that accompanied this software.  Any inquiries         #
# concerning the scope or enforceability of the license should be addressed   #
# to:                                                                         #
#                                                                             #
# Plone Foundation                                                            #
# 4617 Montrose Blvd., Suite C215                                             #
# Houston, Texas 77006 USA                                                    #
# p. +1 713.942.2377 | f. +1 832.201.8856                                     #
# www.enfoldsystems.com                                                       #
# info@enfoldsystems.com                                                      #
#                                                                             #
###############################################################################


from sets import Set
from types import ListType, TupleType
import os, logging

from beatbox import PythonClient as Client

from DateTime import DateTime
from Globals import InitializeClass, package_home
from OFS.Cache import Cacheable
from AccessControl import ClassSecurityInfo
from AccessControl.SecurityManagement import getSecurityManager
from Acquisition import aq_base
from Products.PluggableAuthService.plugins.BasePlugin import BasePlugin
from Products.PluggableAuthService.utils import classImplements
from Products.PluggableAuthService.interfaces import plugins as plugin_ifaces
from Products.CMFCore.permissions import ManagePortal
from Products.PageTemplates.PageTemplateFile import PageTemplateFile
from Products.PlonePAS.sheet import MutablePropertySheet, validateValue
from Products.PlonePAS.interfaces.plugins import IMutablePropertiesPlugin

from config import field_mapping, fullname_format, USERNAME_FIELD, PASSWD_FIELD
from utils import DT2dt

logger = logging.getLogger('SalesforceConnector')

_ptdir = os.path.join(package_home(globals()), 'www')

SalesforceConnectorAddForm = PageTemplateFile(
        'SalesforceConnectorAddForm', _ptdir)
SalesforceConnectorEditForm = PageTemplateFile(
        'SalesforceConnectorEditForm', _ptdir)

def manage_addSalesforceConnector(self, id, title, sfusername, sfpasswd,
        REQUEST=None):
    """ """
    self = self.this()
    plugin = SalesforceConnector(id, title)
    self._setObject(id, plugin)
    plugin = getattr(aq_base(self), id)
    plugin.manage_updateConfig(sfusername, sfpasswd)
    if REQUEST is not None:
        REQUEST.RESPONSE.redirect('%s/manage_main' % self.absolute_url())
    

class SalesforceConnector(BasePlugin, Cacheable):

    security = ClassSecurityInfo()

    meta_type = 'Salesforce Connector'

    manage_options = (
            dict(label='Config', action='manage_configForm'),
            ) + BasePlugin.manage_options + Cacheable.manage_options

    sfusername = ''
    sfpasswd = ''

    security.declareProtected(ManagePortal, 'manage_configForm')
    manage_configForm = SalesforceConnectorEditForm


    def __init__(self, id, title=''):
        """ Initialize a new instance """
        self.id = id
        self.title = title

    def _login(self):
        username = self.sfusername
        passwd = self.sfpasswd
        res = self._getSvc().login(username, passwd)

    _v_svc = None

    def _getSvc(self):
        if self._v_svc is None:
            self._v_svc = Client()
        return self._v_svc


    security.declarePrivate('enumerateUsers')
    def enumerateUsers(self, id=None, login=None, exact_match=False,
        sort_by=None, max_results=None, **kw):
        """ See IUserEnumeration
        """
        self._login()
        ids = list()
        for x in (id, login):
            if x is not None:
                if type(x) in (ListType, TupleType):
                    ids.extend(list(x))
                else:
                    ids.append(x)
        if exact_match:
            query_template = "%s='%s'"
        else:
            query_template = "%s like '%%%s%%'"
        results = list()
        for id in ids:
            if max_results and len(results) >= max_results:
                break
            res = self._getSvc().query(USERNAME_FIELD, 'Contact', query_template % (
                    USERNAME_FIELD, id))
            results.extend(self._extractData(res))
            while not res['done']:
                if max_results and len(results) >= max_results:
                    break
                res = self._getSvc().queryMore(res['queryLocator'])
                results.extend(self._extractData(res))
        if max_results:
            results = results[:max_results]
        return tuple(results)
                

    def _extractData(self, res):
        results = list()
        for r in res['records']:
            username = r[USERNAME_FIELD]
            data = dict(id=username, login=username, 
                pluginid=self.getId())
            results.append(data)
        return results
    
    security.declarePrivate('authenticateCredentials')
    def authenticateCredentials(self, credentials):
        """ See IAuthenticationPlugin
        """
        login = credentials.get('login')
        password = credentials.get('password')
        logger.debug('Calling authenticateCredentials for %s' % login)
        view_name = '%s_authenticateCredentials_%s' % (self.getId(), login)
        keywords=dict(password=password)
        cached_info = self.ZCacheable_get(view_name = view_name, 
            keywords=keywords, default=None)
        if cached_info is not None:
            logger.debug('Returning cached info for '
            'authenticateCredentials for %s' % login)
            return cached_info
        self._login()
        res = self._getSvc().query(USERNAME_FIELD, 'Contact',
            "%s = '%s' and %s = '%s'" % (USERNAME_FIELD, login, 
                PASSWD_FIELD, password))
        if res['size'] == 0:
            result =  None
        else:
            username = res['records'][0][USERNAME_FIELD]
            result = (username, username)
        self.ZCacheable_set(result, view_name=view_name, keywords=keywords)
        return result

    def updateCredentials(self, request, response, login, new_password):
        self._login()
        res = self._getSvc().query('Id', 'Contact', "%s = '%s'" % (
            USERNAME_FIELD, login))
        assert res['size'] == 1
        data = dict(type='Contact', Id=res['records'][0]['Id'])
        data[PASSWD_FIELD] = new_password
        res = self._getSvc().update(data)
        view_name = '%s_authenticateCredentials_%s' % (self.getId(), login)
        self.ZCacheable_invalidate(view_name=view_name)

    def resetCredentials(self, request, response):
        user = getSecurityManager().getUser()
        view_name = '%s_authenticateCredentials_%s' % (self.getId(), 
            user.getId())
        self.ZCacheable_invalidate(view_name=view_name)
        

    def getPropertiesForUser(self, user, request=None):
        view_name = '%s_getPropertiesForUser_%s' % (self.getId(), user.getId())
        cached_info = self.ZCacheable_get(view_name = view_name, default=None)
        if cached_info is not None:
            return cached_info
        self._login()
        schema = list()
        fields = [x['field'] for x in field_mapping.values()]
        fields = ','.join(fields)
        res = self._getSvc().query(fields, 'Contact', "%s='%s'" % (USERNAME_FIELD,
                user.getId()))
        assert res['size'] == 1
        data = dict()
        for fname, fdata in field_mapping.items():
            schema.append((fname, fdata['type']))
            data[fname] = res['records'][0][fdata['field']]
        data['fullname'] = fullname_format % data
        result = MutablePropertySheet(self.id, schema, **data)
        self.ZCacheable_set(result, view_name=view_name)
        return result

    def setPropertiesForUser(self, user, propertysheet):
        view_name = '%s_getPropertiesForUser_%s' % (self.getId(), user.getId())
        self.ZCacheable_invalidate(view_name=view_name)
        self._login()
        properties = dict(propertysheet.propertyItems())
        res = self._getSvc().query('Id', 'Contact', "%s = '%s'" % (USERNAME_FIELD, 
                user.getId()))
        assert res['size'] == 1
        data = dict(type='Contact', Id=res['records'][0]['Id'])
        for fname, fdata in field_mapping.items():
            if (fname in properties and not 
                    validateValue(fdata['type'], properties[fname])):
                raise ValueError, ('Invalid value: %s does not conform '
                        'to %s' % (fname, fdata['type']))
            value = properties[fname]
            if value is None:
                value = field_mapping[fname]['default']
            if isinstance(value, DateTime):
                value = DT2dt(value)
            data[fdata['field']] = value
        unknown = Set(properties.keys()) - Set(field_mapping.keys())
        if unknown:
            raise ValueError, 'Unknown Properties: %r' % unknown
        res = self._getSvc().update(data)
        assert res[0]['success']

    def deleteUser(self, userid):
        raise ValueError, 'Deleting of user data is not allowed.'

    security.declarePrivate('doAddUser')
    def doAddUser(self, login, password):
        self._login()
        res = self._getSvc().query('Id', 'Contact', "%s='%s'" % (USERNAME_FIELD,
                login))
        if res['size'] != 0:
            return False
        data = dict()
        data['type'] = 'Contact'
        data['LastName'] = login
        data[USERNAME_FIELD] = login
        data[PASSWD_FIELD] = password
        res = self._getSvc().create(data)
        if res[0]['success']:
            return True
        return False

    security.declareProtected(ManagePortal, 'manage_updateConfig')
    def manage_updateConfig(self, sfusername, sfpasswd, REQUEST=None):
        """ Update the config """
        self.sfusername = sfusername
        self.sfpasswd = sfpasswd
        if REQUEST is not None:
            REQUEST.RESPONSE.redirect('%s/manage_config' % self.absolute_url())

InitializeClass(SalesforceConnector)


classImplements(SalesforceConnector,
    plugin_ifaces.IUserEnumerationPlugin,
    plugin_ifaces.IAuthenticationPlugin,
    plugin_ifaces.ICredentialsUpdatePlugin,
    #plugin_ifaces.ICredentialsResetPlugin,
    plugin_ifaces.IPropertiesPlugin,
    IMutablePropertiesPlugin,
    plugin_ifaces.IUserAdderPlugin,
    )

